
/* Copyright 2021 Haakan T. Johansson */

/*  This file is part of WIGSGLL.
 *
 *  WIGSGLL is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  WIGSGLL is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with WIGSGLL.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */

#ifndef __WIGSGLL_AVOID_H__
#define __WIGSGLL_AVOID_H__

#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

/***********************************************************************
 *
 * See wigsgll.h for naming scheme.
 */

/* Calculate using long double, return as double.
 *
 * Not recommended!  Slightly better precision than double on x86!
 * (Faster than the ..jjqd.. routines.)
 * But only gives full double precision results up to j = ~10.
 */

int wig3jjldjd  (/*   x   */ int two_j2, int two_j3,
                 /* -m2-m3*/ int two_m2, int two_m3,
                 double *res, size_t entries,
                 int *two_j1_min, int *two_j1_max);

int wig3jjldmd  (int two_j1, int two_j2, int two_j3,
                 int two_m1, /*   x   */ /* -m1-x */
                 double *res, size_t entries,
                 int *two_m2_min, int *two_m2_max);

int wig6jjldjd  (/*   x   */ int two_j2, int two_j3,
                 int two_l1, int two_l2, int two_l3,
                 double *res, size_t entries,
                 int *two_j1_min, int *two_j1_max);

int wig9jjldd   (int two_j1, int two_j2, int two_j3,
		 int two_j4, int two_j5, int two_j6,
		 int two_j7, int two_j8, int two_j9,
		 double *res);

/* Calculate single symbols.  (Not faster!) */

double wig3jjldd(int two_j1, int two_j2, int two_j3,
		 int two_m1, int two_m2, int two_m3);

double wig6jjldd(int two_j1, int two_j2, int two_j3,
		 int two_l1, int two_l2, int two_l3);

/**********************************************************************/

#ifdef __cplusplus
}
#endif

#endif/*__WIGSGLL_AVOID_H__*/
