#!/usr/bin/perl

use strict;
use warnings;

# Get the two file names.
my ($file1, $file2) = @ARGV;

# Open the files.
open my $fh1, "<", $file1 or die $!;
open my $fh2, "<", $file2 or die $!;

my $diff = 0;
my $line1, my $line2;

 compare_lines:
while ($line1 = <$fh1>)
{
    # If we got a line from file1, but none from file2, they differ.
    if (!defined($line2 = <$fh2>))
    {
	print "-$line1";
	$diff = 1;
	last compare_lines;
    }

    # If the files are not a perfrect match, we need to check.
    if ($line1 ne $line2)
    {
	# Split the lines by spaces, parenthesis and equal signs.
	my @obj1 = split(/([\s\(\)=]+)/,$line1);
	my @obj2 = split(/([\s\(\)=]+)/,$line2);

	my $linediff = 0;

	if ($#obj1 != $#obj2)
	{
	    # Different number of objects mean mismatch
	    $linediff = 1;
	}
	else
	{
	    # Check all items in turn.
	    for (my $i = 0; $i <= $#obj1; $i++)
	    {
		# Kill spaces from items, such that separators can
		# differ in amount of spaces.
		$obj1[$i] =~ s/\s//g;
		$obj2[$i] =~ s/\s//g;

		# Check if the items match.
		if ($obj1[$i] ne $obj2[$i])
		{
		    # Are the mismatching items (float) values?
		    if ($obj1[$i] =~ /-?[0-9.]+([eE][+-]?[0-9]+)?/ &&
			$obj2[$i] =~ /-?[0-9.]+([eE][+-]?[0-9]+)?/)
		    {
			my $v1 = $obj1[$i];
			my $v2 = $obj2[$i];
			my $abserr = abs($v1 - $v2);
			my $relerr = $abserr / abs($v1 + $v2);

			# If the relative difference is small,
			# or the absolute difference is, then we accept.
			if ($relerr > 1e-15 && $abserr > 1e-20)
			{
			    print "num: '$obj1[$i]' '$obj2[$i]'\n";
			    $linediff = 1;
			}
		    }
		    else
		    {
			print "'$obj1[$i]' '$obj2[$i]'\n";
			$linediff = 1;
		    }
		}
	    }
	}

	# There was some difference in this line.
	if ($linediff)
	{
	    print "-$line1";
	    print "+$line2";
	    $diff = 1;
	}
    }
}

# Leftover lines mean difference.
while ($line1 = <$fh1>)
{
    print "-$line1";
    $diff = 1;
}

while ($line2 = <$fh2>)
{
    print "+$line2";
    $diff = 1;
}

# Close the files.
close $fh1;
close $fh2;

# Return the overall result.
exit $diff;
