
/* Copyright 2023 Haakan T. Johansson */

/*  This file is part of WIGSGLL.
 *
 *  WIGSGLL is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  WIGSGLL is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with WIGSGLL.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */

RETURN_TYPE FCN_POSTFIX(3,,CALC_SUFFIX,RETURN_SUFFIX)
  (
   int two_j1,  int two_j2,  int two_j3,
   int two_m1,  int two_m2,  int two_m3
   )
{
  int two_j1_min, two_j1_max, n_j1;
  int two_j2_min, two_j2_max, n_j2;
  int two_j3_min, two_j3_max, n_j3;
  int two_m1_min, two_m1_max, n_m1;
  int two_m2_min, two_m2_max, n_m2;
  int two_m3_min, two_m3_max, n_m3;

  int n, kind, off;

  /* To calculate a specific 3j, we can let any of the six arguments
   * be free and calculate the string of symbols associated with that.
   */

  two_j1_min = CHOOSE_MAX(abs(two_j2 - two_j3), abs(two_m1));
  two_j1_max =                two_j2 + two_j3;

  n_j1 = (two_j1_max - two_j1_min) /* / 2 + 1 */;

  /*         */ { n = n_j1; kind = 0; off = two_j1 - two_j1_min; }

  two_j2_min = CHOOSE_MAX(abs(two_j3 - two_j1), abs(two_m2));
  two_j2_max =                two_j3 + two_j1;

  n_j2 = (two_j2_max - two_j2_min) /* / 2 + 1 */;

  if (n_j2 < n) { n = n_j2; kind = 1; off = two_j2 - two_j2_min; }

  two_j3_min = CHOOSE_MAX(abs(two_j1 - two_j2), abs(two_m3));
  two_j3_max =                two_j1 + two_j2;

  n_j3 = (two_j3_max - two_j3_min) /* / 2 + 1 */;

  if (n_j3 < n) { n = n_j3; kind = 2; off = two_j3 - two_j3_min; }

  two_m1_min = CHOOSE_MAX(-two_j1, - two_j2 - two_m3);
  two_m1_max = CHOOSE_MIN( two_j1,   two_j2 - two_m3);

  n_m1 = (two_m1_max - two_m1_min) /* / 2 + 1 */;

  if (n_m1 < n) { n = n_m1; kind = 3; off = two_m1 - two_m1_min; }

  two_m2_min = CHOOSE_MAX(-two_j2, - two_j3 - two_m1);
  two_m2_max = CHOOSE_MIN( two_j2,   two_j3 - two_m1);

  n_m2 = (two_m2_max - two_m2_min) /* / 2 + 1 */;

  if (n_m2 < n) { n = n_m2; kind = 4; off = two_m2 - two_m2_min; }

  two_m3_min = CHOOSE_MAX(-two_j3, - two_j1 - two_m2);
  two_m3_max = CHOOSE_MIN( two_j3,   two_j1 - two_m2);

  n_m3 = (two_m3_max - two_m3_min) /* / 2 + 1 */;

  if (n_m3 < n) { n = n_m3; kind = 5; off = two_m3 - two_m3_min; }

  /* The common operation. */

  n = n / 2 + 1;
  off = off / 2;

  RETURN_TYPE tmp[n];

  /* And do the correct call. */

  switch (kind)
    {
    case 0:
      FCN_POSTFIX(3,j,CALC_SUFFIX,RETURN_SUFFIX)
	(/* x=j1 */  two_j2,  two_j3,
	 /*two_m1*/  two_m2,  two_m3,
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
    case 1:
      FCN_POSTFIX(3,j,CALC_SUFFIX,RETURN_SUFFIX)
	(/* x=j2 */  two_j3,  two_j1,
	 /*two_m2*/  two_m3,  two_m1,
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
    case 2:
      FCN_POSTFIX(3,j,CALC_SUFFIX,RETURN_SUFFIX)
	(/* x=j3 */  two_j1,  two_j2,
	 /*two_m3*/  two_m1,  two_m2,
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
    case 3:
      FCN_POSTFIX(3,j,CALC_SUFFIX,RETURN_SUFFIX)
	(two_j3,  two_j1,     two_j2,
	 two_m3,  /* x=m1 */ /*-m3-x*/
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
    case 4:
      FCN_POSTFIX(3,j,CALC_SUFFIX,RETURN_SUFFIX)
	(two_j1,  two_j2,     two_j3,
	 two_m1,  /* x=m2 */ /*-m1-x*/
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
    case 5:
      FCN_POSTFIX(3,j,CALC_SUFFIX,RETURN_SUFFIX)
	(two_j2,  two_j3,     two_j1,
	 two_m2,  /* x=m3 */ /*-m2-x*/
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
    }

  /*
  printf ("%2d %2d %2d   %2d %2d %2d   "
	  "kind: %d   %2d %2d   %2d %2d   %2d %2d\n",
	  two_j1,  two_j2,  two_j3,
	  two_m1,  two_m2,  two_m3,
	  kind, n_j1, n_m1, n_j2, n_m2, n_j3, n_m3);
  */

  /* printf ("kind: %d\n", kind); */

  RETURN_TYPE res;
  
  /* Pick the wanted value. */
  res = tmp[off];

  return res;
}
