
/* Copyright 2023 Haakan T. Johansson */

/*  This file is part of WIGSGLL.
 *
 *  WIGSGLL is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  WIGSGLL is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with WIGSGLL.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */

RETURN_TYPE FCN_POSTFIX(6,,CALC_SUFFIX,RETURN_SUFFIX)
  (
   int two_j1,  int two_j2,  int two_j3,
   int two_l1,  int two_l2,  int two_l3
   )
{
  int two_j1_min, two_j1_max, n_j1;
  int two_j2_min, two_j2_max, n_j2;
  int two_j3_min, two_j3_max, n_j3;
#if 0
  int two_l1_min, two_l1_max, n_l1;
  int two_l2_min, two_l2_max, n_l2;
  int two_l3_min, two_l3_max, n_l3;
#endif

  int n, kind, off;

  /* To calculate a specific 6j, we can let any of the six arguments
   * be free and calculate the string of symbols associated with that.
   *
   * The range limits are given by the four arguments in the other
   * columns.  When a j is chosen as free, the limits are given by
   * the j and l pairs.  When a l is chosen as free, the limits
   * are given by mixed j-l pairs.
   *
   * Testing shows that l free give no new opportunities (always
   * the same number of items as the corrsponding j cases).
   */

  two_j1_min = CHOOSE_MAX(abs(two_j2 - two_j3), abs(two_l2 - two_l3));
  two_j1_max = CHOOSE_MIN(    two_j2 + two_j3 ,     two_l2 + two_l3 );

  n_j1 = (two_j1_max - two_j1_min) /* / 2 + 1 */;

  /*         */ { n = n_j1; kind = 0; off = two_j1 - two_j1_min; }

  two_j2_min = CHOOSE_MAX(abs(two_j1 - two_j3), abs(two_l1 - two_l3));
  two_j2_max = CHOOSE_MIN(    two_j1 + two_j3 ,     two_l1 + two_l3 );

  n_j2 = (two_j2_max - two_j2_min) /* / 2 + 1 */;

  if (n_j2 < n) { n = n_j2; kind = 1; off = two_j2 - two_j2_min; }

  two_j3_min = CHOOSE_MAX(abs(two_j2 - two_j1), abs(two_l2 - two_l1));
  two_j3_max = CHOOSE_MIN(    two_j2 + two_j1 ,     two_l2 + two_l1 );

  n_j3 = (two_j3_max - two_j3_min) /* / 2 + 1 */;

  if (n_j3 < n) { n = n_j3; kind = 2; off = two_j3 - two_j3_min; }

#if 0
  two_l1_min = CHOOSE_MAX(abs(two_l2 - two_j3), abs(two_j2 - two_l3));
  two_l1_max = CHOOSE_MIN(    two_l2 + two_j3 ,     two_j2 + two_l3 );

  n_l1 = (two_l1_max - two_l1_min) /* / 2 + 1 */;

  if (n_l1 < n) { n = n_l1; kind = 3; off = two_l1 - two_l1_min; }

  two_l2_min = CHOOSE_MAX(abs(two_l1 - two_j3), abs(two_j1 - two_l3));
  two_l2_max = CHOOSE_MIN(    two_l1 + two_j3 ,     two_j1 + two_l3 );

  n_l2 = (two_l2_max - two_l2_min) /* / 2 + 1 */;

  if (n_l2 < n) { n = n_l2; kind = 4; off = two_l2 - two_l2_min; }

  two_l3_min = CHOOSE_MAX(abs(two_l2 - two_j1), abs(two_j2 - two_l1));
  two_l3_max = CHOOSE_MIN(    two_l2 + two_j1 ,     two_j2 + two_l1 );

  n_l3 = (two_l3_max - two_l3_min) /* / 2 + 1 */;

  if (n_l3 < n) { n = n_l3; kind = 5; off = two_l3 - two_l3_min; }
#endif

  /* The common operation. */

  n = n / 2 + 1;
  off = off / 2;

  RETURN_TYPE tmp[n];

  /* And do the correct call. */

  switch (kind)
    {
    case 0:
      FCN_POSTFIX(6,j,CALC_SUFFIX,RETURN_SUFFIX)
	(/* x=j1 */  two_j2,  two_j3,
	 two_l1,     two_l2,  two_l3,
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
    case 1:
      FCN_POSTFIX(6,j,CALC_SUFFIX,RETURN_SUFFIX)
	(/* x=j2 */  two_j3,  two_j1,
	 two_l2,     two_l3,  two_l1,
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
    case 2:
      FCN_POSTFIX(6,j,CALC_SUFFIX,RETURN_SUFFIX)
	(/* x=j3 */  two_j1,  two_j2,
	 two_l3,     two_l1,  two_l2,
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
#if 0
    case 3:
      FCN_POSTFIX(6,j,CALC_SUFFIX,RETURN_SUFFIX)
	(/* x=l1 */  two_l2,  two_j3,
	 two_j1,     two_j2,  two_l3,
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
    case 4:
      FCN_POSTFIX(6,j,CALC_SUFFIX,RETURN_SUFFIX)
	(/* x=l2 */  two_l3,  two_j1,
	 two_j2,     two_j3,  two_l1,
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
    case 5:
      FCN_POSTFIX(6,j,CALC_SUFFIX,RETURN_SUFFIX)
	(/* x=l3 */  two_l1,  two_j2,
	 two_j3,     two_j1,  two_l2,
	 tmp, (size_t) n,
	 NULL, NULL);
      break;
#endif
    }

  /*
  printf ("%2d %2d %2d   %2d %2d %2d   "
	  "kind: %d   %2d %2d   %2d %2d   %2d %2d\n",
	  two_j1,  two_j2,  two_j3,
	  two_l1,  two_l2,  two_l3,
	  kind, n_j1, n_l1, n_j2, n_l2, n_j3, n_l3);
  */

  RETURN_TYPE res;
  
  /* Pick the wanted value. */
  res = tmp[off];

  return res;
}
