
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include <quadmath.h>
#include <string.h>

#include "wigsgll.h"

#include "wigxjpf.h"
#include "wigxjpf_quadmath.h"

#include "calc_minmax.h"

__float128 vect1[3][0x10000];
__float128 vect2[0x10000];
__float128 vect3[0x10000];

__float128 v_maxabsrelerr[0x10000][3];
__float128 v_maxabserrz[0x10000][3];
__float128 v_maxabserr[0x10000][3];
int        v_num[0x10000];

/* Routines used for tests, that are not included in the normal setup
 * of wigsgll.h, since they calculate with lower precision than they
 * return result with.
 *
 * (This way we can use same array for all routines.)
 */

int wig3jjdjqd  (/*   x   */ int two_j2, int two_j3,
                 /* -m2-m3*/ int two_m2, int two_m3,
                 __float128 *res, size_t entries,
                 int *two_j1_min, int *two_j1_max);

int wig3jjdmqd  (int two_j1, int two_j2, int two_j3,
                 int two_m1, /*   x   */ /* -m1-x */
                 __float128 *res, size_t entries,
                 int *two_m2_min, int *two_m2_max);

int wig6jjdjqd  (/*   x   */ int two_j2, int two_j3,
                 int two_l1, int two_l2, int two_l3,
                 __float128 *res, size_t entries,
                 int *two_j1_min, int *two_j1_max);

__float128 wig3jjdqd   (int two_j1, int two_j2, int two_j3,
			int two_m1, int two_m2, int two_m3);

__float128 wig6jjdqd   (int two_j1, int two_j2, int two_j3,
			int two_l1, int two_l2, int two_l3);

int wig9jjdqd   (int two_j1, int two_j2, int two_j3,
		 int two_j4, int two_j5, int two_j6,
		 int two_j7, int two_j8, int two_j9,
		 __float128 *res);

/* */

int wig3jjldjqd (/*   x   */ int two_j2, int two_j3,
                 /* -m2-m3*/ int two_m2, int two_m3,
                 __float128 *res, size_t entries,
                 int *two_j1_min, int *two_j1_max);

int wig3jjldmqd (int two_j1, int two_j2, int two_j3,
                 int two_m1, /*   x   */ /* -m1-x */
                 __float128 *res, size_t entries,
                 int *two_m2_min, int *two_m2_max);

int wig6jjldjqd (/*   x   */ int two_j2, int two_j3,
                 int two_l1, int two_l2, int two_l3,
                 __float128 *res, size_t entries,
                 int *two_j1_min, int *two_j1_max);

int wig9jjldqd  (int two_j1, int two_j2, int two_j3,
		 int two_j4, int two_j5, int two_j6,
		 int two_j7, int two_j8, int two_j9,
		 __float128 *res);

/* */

double elapsed;

#define ELAPSE(b,a) ((double) (a.tv_sec - b.tv_sec) +			\
		     0.000001 * (double) (a.tv_usec - b.tv_usec))

#define TEST_6J 1
#define TEST_3J 0

typedef struct test_36jm_t
{
  /*     */   int two_j2; int two_j3;
  int two_l1; int two_l2; int two_l3;
  /* For 9j only: */
  /*     */               int two_j6;
  int two_j7; int two_j8; int two_j9;
} test_36jm;

/* Do it such that l2, l3 are m, and l1 as j.
 * Such that random can do negative range for the m.
 */
/* For 3j-j. */
#define two_m2 two_l2
#define two_m3 two_l3
/* For 3j-m. */
#define two_j1 two_l1
#define two_m1 two_l2
/* For 9j.   */
#define two_j4 two_l2
#define two_j5 two_l3


typedef struct test_config_t
{
  int do_d;
  int do_ld;
  int do_qd;
  int do_3j_j;
  int do_3j_m;
  int do_6j_j;
  int do_9j;
  int two_lim_min;
  int two_lim_max;
  int random;
  int timing;
  int show_progress;
} test_config;

test_config config;

int had_mismatch = 0;

int interestingx = 0;

int test(test_36jm *t)
{
  int two_x_min, two_x_max;
  int two_x;

  int ret = 1;
  int i;
  int dldqd;

#define two_xm2_min  two_x_min
#define two_xm2_max  two_x_max
#define two_xm2      two_x

#define two_xj1_min  two_x_min
#define two_xj1_max  two_x_max
#define two_xj1      two_x

  __float128 maxabsrelerr = 0;
  __float128 maxabserrz = 0;
  __float128 maxabserr = 0;

  struct timeval a, b;
  struct timeval c, d;
  /* struct timeval e, f; */

  interestingx = 0;

  gettimeofday(&a, NULL);

  for (i = config.timing ? 1000 : 1; i > 0; i--)
    {
  if (config.do_6j_j)
    {
      if (config.do_d)
	ret = wig6jjdjqd (/*      */ t->two_j2, t->two_j3,
			  t->two_l1, t->two_l2, t->two_l3,
			  vect1[0], sizeof (vect1[0]) / sizeof (vect1[0][0]),
			  &two_xj1_min, &two_xj1_max);
      if (config.do_ld)
	ret = wig6jjldjqd(/*      */ t->two_j2, t->two_j3,
			  t->two_l1, t->two_l2, t->two_l3,
			  vect1[1], sizeof (vect1[1]) / sizeof (vect1[1][0]),
			  &two_xj1_min, &two_xj1_max);
      if (config.do_qd)
	ret = wig6jjqdjqd(/*      */ t->two_j2, t->two_j3,
			  t->two_l1, t->two_l2, t->two_l3,
			  vect1[2], sizeof (vect1[2]) / sizeof (vect1[2][0]),
			  &two_xj1_min, &two_xj1_max);
    }
  if (config.do_3j_j)
    {
      if (config.do_d)
	ret = wig3jjdjqd (/*      */ t->two_j2, t->two_j3,
			  /*      */ t->two_m2, t->two_m3,
			  vect1[0], sizeof (vect1[0]) / sizeof (vect1[0][0]),
			  &two_xj1_min, &two_xj1_max);
      if (config.do_ld)
	ret = wig3jjldjqd(/*      */ t->two_j2, t->two_j3,
			  /*      */ t->two_m2, t->two_m3,
			  vect1[1], sizeof (vect1[1]) / sizeof (vect1[1][0]),
			  &two_xj1_min, &two_xj1_max);
      if (config.do_qd)
	ret = wig3jjqdjqd(/*      */ t->two_j2, t->two_j3,
			  /*      */ t->two_m2, t->two_m3,
			  vect1[2], sizeof (vect1[2]) / sizeof (vect1[2][0]),
			  &two_xj1_min, &two_xj1_max);
    }
  if (config.do_3j_m)
    {
      if (config.do_d)
	ret = wig3jjdmqd (t->two_j1, t->two_j2, t->two_j3,
			  t->two_m1, /*      */ /*      */
			  vect1[0], sizeof (vect1[0]) / sizeof (vect1[0][0]),
			  &two_xm2_min, &two_xm2_max);
      if (config.do_ld)
	ret = wig3jjldmqd(t->two_j1, t->two_j2, t->two_j3,
			  t->two_m1, /*      */ /*      */
			  vect1[1], sizeof (vect1[1]) / sizeof (vect1[1][0]),
			  &two_xm2_min, &two_xm2_max);
      if (config.do_qd)
	ret = wig3jjqdmqd(t->two_j1, t->two_j2, t->two_j3,
			  t->two_m1, /*      */ /*      */
			  vect1[2], sizeof (vect1[2]) / sizeof (vect1[2][0]),
			  &two_xm2_min, &two_xm2_max);
    }
  if (config.do_9j)
    {
      if (config.do_d)
	ret = wig9jjdqd (t->two_j1, t->two_j2, t->two_j3,
			 t->two_j4, t->two_j5, t->two_j6,
			 t->two_j7, t->two_j8, t->two_j9,
			 &vect1[0][0]);
      if (config.do_ld)
	ret = wig9jjldqd(t->two_j1, t->two_j2, t->two_j3,
			 t->two_j4, t->two_j5, t->two_j6,
			 t->two_j7, t->two_j8, t->two_j9,
			 &vect1[1][0]);
      if (config.do_qd)
	ret = wig9jjqdqd(t->two_j1, t->two_j2, t->two_j3,
			 t->two_j4, t->two_j5, t->two_j6,
			 t->two_j7, t->two_j8, t->two_j9,
			 &vect1[2][0]);
      ret = 1;
      two_x_min = 0;
      two_x_max = 0;
    }
    }
  gettimeofday(&b, NULL);

  if (ret == 0)
    {
      return ret;
    }

  gettimeofday(&c, NULL);

  /* Calculate reference values with WIGXJPF. */

  for (i = 0, two_x = two_x_min; two_x <= two_x_max; two_x += 2, i++)
    {
      if (config.do_6j_j)
	{
	  wig6jj_float128(&vect2[i],
			  two_xj1,   t->two_j2, t->two_j3,
			  t->two_l1, t->two_l2, t->two_l3);
	}
      if (config.do_3j_j)
	{
	  int two_m1 = - t->two_m2 - t->two_m3;

	  wig3jj_float128(&vect2[i],
			  two_xj1, t->two_j2, t->two_j3,
			  two_m1,  t->two_m2, t->two_m3);
	}
      if (config.do_3j_m)
	{
	  int two_m3 = - t->two_m1 - two_xm2;

	  wig3jj_float128(&vect2[i],
			  t->two_j1, t->two_j2, t->two_j3,
			  t->two_m1, two_xm2,   two_m3   );
	}
      if (config.do_9j)
	{
	  wig9jj_float128(&vect2[i],
			  t->two_j1, t->two_j2, t->two_j3,
			  t->two_j4, t->two_j5, t->two_j6,
			  t->two_j7, t->two_j8, t->two_j9);
	}

      /*
      if (vect2[i] == 0.0)
	printf ("                                Zero!\n");
      */
    }

  gettimeofday(&d, NULL);

  int two_max_j = 0;

  if (t->two_j2 > two_max_j)     two_max_j = t->two_j2;
  if (t->two_j3 > two_max_j)     two_max_j = t->two_j3;
  if (config.do_3j_m)
    {
      if (t->two_j1 > two_max_j)     two_max_j = t->two_j1;
    }
  if (config.do_6j_j)
    {
      if (t->two_l1 > two_max_j)     two_max_j = t->two_l1;
      if (t->two_l2 > two_max_j)     two_max_j = t->two_l2;
      if (t->two_l3 > two_max_j)     two_max_j = t->two_l3;
    }
  if (config.do_9j)
    {
      if (t->two_j1 > two_max_j)     two_max_j = t->two_j1;
      if (t->two_j4 > two_max_j)     two_max_j = t->two_j4;
      if (t->two_j5 > two_max_j)     two_max_j = t->two_j5;
      if (t->two_j6 > two_max_j)     two_max_j = t->two_j6;
      if (t->two_j7 > two_max_j)     two_max_j = t->two_j7;
      if (t->two_j8 > two_max_j)     two_max_j = t->two_j8;
      if (t->two_j9 > two_max_j)     two_max_j = t->two_j9;

      /*      
      printf ("%d %d %d  %d %d %d  %d %d %d  %10.15f %20.15f\n",
	      t->two_j1, t->two_j2, t->two_j3,
	      t->two_j4, t->two_j5, t->two_j6,
	      t->two_j7, t->two_j8, t->two_j9,
	      (double) vect1[0][0],
	      (double) vect2[0]);
      */
    }

  for (dldqd = 0; dldqd < 3; dldqd++)
    if ((dldqd == 0 && config.do_d ) ||
	(dldqd == 1 && config.do_ld) ||
	(dldqd == 2 && config.do_qd))
    for (i = 0, two_x = two_x_min; two_x <= two_x_max; two_x += 2, i++)
      {
	__float128 err    = vect1[dldqd][i] - vect2[i];
	__float128 relerr = err / vect2[i];
      
	__float128 abserr    = fabsq(err);
	__float128 absrelerr = fabsq(relerr);

	/* printf ("   %d\n", two_xj1); */

	if (config.do_3j_j || config.do_6j_j)
	  {
	    if (two_xj1 > two_max_j) two_max_j = two_xj1;
	  }

	int v_i = two_max_j;
	/*
	  v_i = ret;
	  v_i = (two_max_j + t->two_j2 + t->two_j3 +
	         t->two_l1 + t->two_l2 + t->two_l3);
	*/

	if (vect2[i] == 0)
	  {
	    if (abserr > maxabserrz)
	      maxabserrz = abserr;
	    if (abserr > v_maxabserrz[v_i][dldqd])
	      v_maxabserrz[v_i][dldqd] = abserr;
	  }
	else
	  {
	    if (absrelerr > maxabsrelerr)
	      maxabsrelerr = absrelerr;
	    if (absrelerr > v_maxabsrelerr[v_i][dldqd])
	      v_maxabsrelerr[v_i][dldqd] = absrelerr;
	  }

	__float128 abserrcorr = abserr * sqrtq(two_max_j+1);
      
	if (abserrcorr > maxabserr)
	  maxabserr = abserrcorr;
	if (abserr > v_maxabserr[v_i][dldqd])
	  v_maxabserr[v_i][dldqd] = abserr;
	v_num[v_i]++;

      }

  double t1 = ELAPSE(a, b);
  double t2 = ELAPSE(c, d);
  /* double t3 = ELAPSE(e, f); */
  double t3 = 0;

  elapsed = t1;

  if (maxabsrelerr > 1.e-8 ||
      maxabserr    > 1.e-10)
    {
      had_mismatch = 1;

      fprintf (stderr,
	       "%d: %4d %4d , %.6f %.6f %.6f  %6.2f  %6.2f  %6.2f    **** ERR\n",
	       ret, two_x_min, two_x_max,
	       t1, t2, t3,
	       (double) log10q(maxabserr),
	       (double) log10q(maxabserrz),
	       (double) log10q(maxabsrelerr));
    }
  return ret;
}

#define DBL_TO_TWO(dest,src) do {			\
    dest = (int) ((src) * 2.);				\
    if (dest != (src) * 2.) {				\
      fprintf (stderr, "Given j (%f) in option '%s' "   \
               "is neither integer nor "                \
               "half-integer.\n", src, argv[i]);	\
      exit(1);                                          \
    }                                                   \
  }                                                     \
  while (0)

int main(int argc, char *argv[])
{
  test_36jm t;
  int j; (void) j;

  int badopt;
  int i;
  int nonzero = 0;

  memset (&config, 0, sizeof(config));
  config.show_progress = 1;
  
  for (i = 1; i < argc; i++)
    {
      badopt = 0;
      
      if (strncmp(argv[i],"--lim=",6) == 0)
        {
	  double lim1, lim2;

	  if (sscanf(argv[i]+6,"%lf:%lf",
			  &lim1,&lim2) == 2)
	    {
	      DBL_TO_TWO(config.two_lim_min, lim1);
	      DBL_TO_TWO(config.two_lim_max, lim2);
	    }
	  else if (sscanf(argv[i]+6,"%lf",
                     &lim1) == 1)
	    {
	      DBL_TO_TWO(config.two_lim_max, lim1);
	    }
	  else
	    badopt = 1;
	}
      else if (strcmp(argv[i],"--random") == 0)
	{
	  config.random = 1;
	}
      else if (strcmp(argv[i],"--plain") == 0)
	{
	  config.random = 0;
	}
      else if (strcmp(argv[i],"--3j-j") == 0)
	{
	  config.do_3j_j = 1;
	}
      else if (strcmp(argv[i],"--3j-m") == 0)
	{
	  config.do_3j_m = 1;
	}
      else if (strcmp(argv[i],"--6j-j") == 0)
	{
	  config.do_6j_j = 1;
	}
      else if (strcmp(argv[i],"--9j") == 0)
	{
	  config.do_9j = 1;
	}
      else if (strcmp(argv[i],"--d") == 0)
	{
	  config.do_d = 1;
	}
      else if (strcmp(argv[i],"--ld") == 0)
	{
	  config.do_ld = 1;
	}
      else if (strcmp(argv[i],"--qd") == 0)
	{
	  config.do_qd = 1;
	}
      else if (strcmp(argv[i],"--timing") == 0)
	{
	  config.timing = 1;
	}
      else if (strcmp(argv[i],"--no-progress") == 0)
	{
	  config.show_progress = 0;
	}
      else
	badopt = 1;

      if (badopt)
        {
          fprintf (stderr, "Unknown option '%s'.\n", argv[i]);
          exit (1);
        }
    }

  if (config.do_3j_j + config.do_3j_m + config.do_6j_j + config.do_9j  != 1)
    {
      fprintf (stderr, "Need one of --3j-j, --3j-m, --6j-j, or --9j.\n");
      exit (1);
    }

  if (config.do_d + config.do_ld + config.do_qd == 0)
    {
      config.do_d  = 1;
      config.do_ld = 1;
      config.do_qd = 1;
    }
  
  memset(v_maxabsrelerr, 0, sizeof (v_maxabsrelerr));
  memset(v_maxabserrz,   0, sizeof (v_maxabserrz));
  memset(v_maxabserr,    0, sizeof (v_maxabserr));
  memset(v_num,          0, sizeof (v_num));

  /* printf ("two_lim_max: %d\n", config.two_lim_max); */

  wig_table_init(config.two_lim_max*2,9);
  wig_temp_init(config.two_lim_max*2);

  if (config.timing)
    {
      int mul;
      time_t a, b;

      a = time(NULL);

      for ( ; ; )
	{
	  b = time(NULL);
	  if (b > a + 1 || b < a)
	    break;
	}
      
      
      for (mul = 1; mul < config.two_lim_max; mul *= 2)
	{
	  int ret;

	  if (config.do_6j_j)
	    {
	      t.two_j2 =  mul*1;
	      t.two_j3 =  mul*1;
	      t.two_l1 = (mul*1) & ~1;
	      t.two_l2 =  mul*1;
	      t.two_l3 = 0;
	    }
	  
	  if (config.do_3j_j)
	    {
	      t.two_j2 =  mul*1;
	      t.two_j3 =  mul*1;
	      t.two_m2 =  mul*1;
	      t.two_m3 =  mul*1;
	    }
	  
	  if (config.do_3j_m)
	    {
	      t.two_j1 =  mul*1;
	      t.two_j2 = 0;
	      t.two_j3 =  mul*1;
	      t.two_m1 =  mul*1;
	    }
	  
	  ret = test(&t);

	  printf ("%4d %9.1f %5.1f  ",
		  ret, 1.e6 * elapsed, 1.e6 * elapsed / ret);

	  if (config.do_6j_j)
	    {
	      t.two_j2 =  mul*1;
	      t.two_j3 =  mul*1;
	      t.two_l1 = (mul*1) & ~1;
	      t.two_l2 =  mul*1;
	      t.two_l3 =  mul*1;
	    }
	  
	  if (config.do_3j_j)
	    {
	      t.two_j2 =  mul*1;
	      t.two_j3 =  mul*1;
	      t.two_m2 =  mul*1;
	      t.two_m3 = -mul*1;
	    }
	  
	  if (config.do_3j_m)
	    {
	      t.two_j1 =  mul*1;
	      t.two_j2 =  mul*1;
	      t.two_j3 =  mul*1;
	      t.two_m1 =  mul*1;
	    }
	  
	  ret = test(&t);

	  printf ("%4d %9.1f %5.1f\n",
		  ret, 1.e6 * elapsed, 1.e6 * elapsed / ret);
	}

      exit (0);
    }

  if (!config.random &&
      (config.do_3j_j || config.do_6j_j))
    {
      /* Can interchange any columns, so we can assume j2 >= j3.
       */

#define FOR_LOOP(var, min, max, incr)			\
      for (var = (min); var <= (max); var += (incr))

      FOR_LOOP(t.two_j2, config.two_lim_min, config.two_lim_max, 1) {
	FOR_LOOP(t.two_j3, 0, t.two_j2, 1)
	  {
	    if (config.show_progress)
	      {
		fprintf (stderr, "%d %d\r", t.two_j2, t.two_j3);
		fflush (stderr);
	      }

	    if (config.do_6j_j)
	      {
		/* When testing 6j, we can always say that j2 >= j3,
		 * and j2 >= l2 and j2 >= l3, i.e. j2 is the largest of
		 * j2, j3, l2 and l3.
		 *
		 * Furthermore, l3 must be even/odd such that j2+j3 and
		 * l2+l3 have same even/odd.
		 */

		FOR_LOOP(t.two_l2, 0, t.two_j2, 1)
		  {
		    int l3_start = (t.two_j2 + t.two_j3 + t.two_l2) & 1;
		  
		    FOR_LOOP(t.two_l3, l3_start, t.two_j2, 2)
		      {
			int two_l1_min = CHOOSE_MAX(abs(t.two_j2 - t.two_j3),
						    abs(t.two_l2 - t.two_l3));
			int two_l1_max = CHOOSE_MIN(    t.two_j2 + t.two_j3 ,
							t.two_l2 + t.two_l3 );

			FOR_LOOP(t.two_l1, two_l1_min, two_l1_max, 2)
			  {
			    test(&t);
			  }
		      }
		  }
	      }

	    if (config.do_3j_j)
	      {
		FOR_LOOP(t.two_m2, -t.two_j2, t.two_j2, 2)
		  FOR_LOOP(t.two_m3, -t.two_j3, t.two_j3, 2)
		  {
		    test(&t);
		  }
	      }
	  }
      }
    }

  if (!config.random &&
      config.do_3j_m)
    {
      /* Can interchange any columns, so can assume j1 >= j3.
       */

      FOR_LOOP(t.two_j1, config.two_lim_min, config.two_lim_max, 1) {
	FOR_LOOP(t.two_j3, 0, t.two_j1, 1)
	  {
	    if (config.show_progress)
	      {
		fprintf (stderr, "%d %d\r", t.two_j1, t.two_j3);
		fflush (stderr);
	      }

	    /* j2 is limited by possible values given by j1 and j3,
	     * and even/odd as well.
	     *
	     * We do _not_ limit j2 by the config limit, since
	     * if we do incremental runs of config limit, we would
	     * then miss symbols later.
	     */

	    int two_j2_min = abs(t.two_j1 - t.two_j3);
	    int two_j2_max =     t.two_j1 + t.two_j3;
	    		
	    FOR_LOOP(t.two_j2, two_j2_min, two_j2_max, 2) {
	      FOR_LOOP(t.two_m1, -t.two_j1, t.two_j1, 2)
		{
		  test(&t);
		}
	    }
	  }
      }
    }


  if (!config.random &&
      config.do_9j)
    {
      /* Can interchange any columns, and transpose, so also interchange
       * any row.
       *
       * j1 j2 j3
       * j4 j5 j6
       * j7 j8 j9
       *
       * We can therefore start out with j1 being the largest of all
       * elements, and j2 being second largest.  j1 >= j2, j2 >= j3.
       *
       * Neither j4 nor j7 can be larger than j2, as they would then
       * have been selected for movement to j2.
       *
       * j5, j6, j8, j9 cannot be larger than j1.  But may be larger
       * than j2, since they cannot be moved to that position.
       *
       * For j3, j6, j7, j8 triangle conditions apply.
       * For j9 two triangle conditions apply.
       */

#define TRIANGLE_OK(a,b,c) ((t.two_j##c >= abs(t.two_j##a - t.two_j##b)) && \
			    (t.two_j##c <=    (t.two_j##a + t.two_j##b)) && \
			    !((t.two_j##c ^ t.two_j##a ^ t.two_j##b) & 1))
      
      FOR_LOOP(t.two_j1, config.two_lim_min, config.two_lim_max, 1) {
	FOR_LOOP(t.two_j2, 0, t.two_j1, 1) {
	  int two_j3_min = abs(t.two_j1 - t.two_j2);
	  FOR_LOOP(t.two_j3, two_j3_min, t.two_j2, 2) {
	    if (config.show_progress)
	      {
		fprintf (stderr, "%d %d %d  \r",
			 t.two_j1, t.two_j2, t.two_j3);
		fflush (stderr);
	      }
	    
	    FOR_LOOP(t.two_j4, 0, t.two_j2, 1) {
	      int two_j7_min =        abs(t.two_j1 - t.two_j4);
	      int two_j7_max = CHOOSE_MIN(t.two_j1 + t.two_j4,
					  t.two_j2);
	      FOR_LOOP(t.two_j7, two_j7_min, two_j7_max, 2) {
		FOR_LOOP(t.two_j5, 0, t.two_j1, 1) {
		  int two_j6_min =        abs(t.two_j4 - t.two_j5);
		  int two_j6_max = CHOOSE_MIN(t.two_j4 + t.two_j5,
					      t.two_j1);
		  FOR_LOOP(t.two_j6, two_j6_min, two_j6_max, 2) {
		    int two_j8_min =        abs(t.two_j2 - t.two_j5);
		    int two_j8_max = CHOOSE_MIN(t.two_j2 + t.two_j5,
						t.two_j1);
		    FOR_LOOP(t.two_j8, two_j8_min, two_j8_max, 2) {
		      int two_j9_min = CHOOSE_MAX(abs(t.two_j3 - t.two_j6),
						  abs(t.two_j7 - t.two_j8));
		      int two_j9_max = min3(          t.two_j3 + t.two_j6,
						      t.two_j7 + t.two_j8,
						      t.two_j1);
		      FOR_LOOP(t.two_j9, two_j9_min, two_j9_max, 2)
			{
			  test(&t);
			}
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }

  if (config.random)
    {
      if (!config.do_9j)
	config.two_lim_max *= 2;
      
      for (j = 0; j < 2000000000ll; j++)
	{
	  int ret;

	  if (j % 1000000 == 0 && config.show_progress)
	    fprintf (stderr, "%d %d\n", j, nonzero);
      
	  t.two_j2 = rand() % (config.two_lim_max + 1);
	  t.two_j3 = rand() % (config.two_lim_max + 1);
	  t.two_l1 = rand() % (config.two_lim_max + 1);
	  if (config.do_6j_j || config.do_9j)
	    {
	      t.two_l2 = rand() % (config.two_lim_max + 1);
	      t.two_l3 = rand() % (config.two_lim_max + 1);
	    }
	  else
	    {
	      t.two_l2 = config.two_lim_max - rand() %
		(2 * config.two_lim_max + 1);
	      t.two_l3 = config.two_lim_max - rand() %
		(2 * config.two_lim_max + 1);
	    }
	  if (config.do_9j)
	    {
	      t.two_j6 = rand() % (config.two_lim_max + 1);
	      t.two_j7 = rand() % (config.two_lim_max + 1);
	      t.two_j8 = rand() % (config.two_lim_max + 1);
	      t.two_j9 = rand() % (config.two_lim_max + 1);

	      if (!(TRIANGLE_OK(1,2,3) &&
		    TRIANGLE_OK(4,5,6) &&
		    TRIANGLE_OK(1,4,7) &&
		    TRIANGLE_OK(2,5,8) &&
		    TRIANGLE_OK(3,6,9) &&
		    TRIANGLE_OK(7,8,9)))
		continue;
     	    }
	  /*	  
	  printf ("%d %d %d  %d %d %d  %d %d %d\n",
		  t.two_j1, t.two_j2, t.two_j3,
		  t.two_j4, t.two_j5, t.two_j6,
		  t.two_j7, t.two_j8, t.two_j9);
	  */
	  ret = test(&t);

	  nonzero += !!ret;
	}
    }

  wig_temp_free();
  wig_table_free();

  for (int i = 0; i <= config.two_lim_max*2; i++)
    {
      if (v_num[i])
	{
	  int dldqd;

	  printf ("%3d", i);
	  for (dldqd = 0; dldqd < 3; dldqd++)
	    printf ("   %6.2f %6.2f %6.2f",
		    (double) log10q(v_maxabserr[i][dldqd]),
		    (double) log10q(v_maxabserrz[i][dldqd]),
		    (double) log10q(v_maxabsrelerr[i][dldqd]));
	  printf ("  %d\n", v_num[i]);
	}
    }

  if (had_mismatch)
    {
      fprintf (stderr, "Had mismatch!\n");
      return 1;
    }

  return 0;
}

/* 
Timing:

3j-j:  d   20 ns   20 ns/sym
3j-j: ld   32 ns   29 ns/sym
3j-j: qd  363 ns  600 ns/sym

3j-m:  d   26 ns   17 ns/sym
3j-m: ld   30 ns   30 ns/sym
3j-m: qd  380 ns  500 ns/sym

6j-j:  d   20 ns   20 ns/sym
6j-j: ld   32 ns   27 ns/sym
6j-j: qd  310 ns  510 ns/sym

*/
